@@ -1,12 +1,12 @@
 use std::collections::HashSet;
 use std::fs::File;
-use std::io::{BufReader, BufWriter};
+use std::io::{BufReader, BufWriter, Cursor};
+use std::path::{Path, PathBuf};
 use std::sync::{Arc, RwLock};
 
 use finalfusion::compat::floret::ReadFloretText;
 use finalfusion::compat::text::{ReadText, ReadTextDims};
 use finalfusion::compat::word2vec::ReadWord2Vec;
-use finalfusion::io::WriteEmbeddings;
 use finalfusion::metadata::Metadata;
 use finalfusion::prelude::*;
 use finalfusion::similarity::*;
@@ -16,10 +16,10 @@ use itertools::Itertools;
 use ndarray::Array1;
 use numpy::{IntoPyArray, PyArray1, PyArray2, PyReadonlyArray1};
 use pyo3::class::iter::PyIterProtocol;
+use pyo3::exceptions;
 use pyo3::exceptions::PyValueError;
 use pyo3::prelude::*;
-use pyo3::types::{PyAny, PyIterator, PyTuple};
-use pyo3::{exceptions, PyMappingProtocol};
+use pyo3::types::{PyAny, PyBytes, PyIterator, PyTuple};
 use reductive::pq::Pq;
 #[cfg(feature = "opq")]
 use reductive::pq::{GaussianOpq, Opq};
@@ -36,6 +36,9 @@ pub struct PyEmbeddings {
 
 #[pymethods]
 impl PyEmbeddings {
+    /// new(path, mmap=False)
+    /// --
+    ///
     /// Load embeddings from the given `path`.
     ///
     /// When the `mmap` argument is `True`, the embedding matrix is
@@ -44,13 +47,13 @@ impl PyEmbeddings {
     /// some query efficiency.
     #[new]
     #[args(mmap = false)]
-    fn new(path: &str, mmap: bool) -> PyResult<PyEmbeddings> {
+    fn new(path: PathBuf, mmap: bool) -> PyResult<PyEmbeddings> {
         // First try to load embeddings with viewable storage. If that
         // fails, attempt to load the embeddings as non-viewable
         // storage.
-        let embeddings = match read_embeddings(path, mmap) {
+        let embeddings = match read_embeddings(&path, mmap) {
             Ok(e) => EmbeddingsWrap::View(e),
-            Err(_) => read_embeddings(path, mmap)
+            Err(_) => read_embeddings(&path, mmap)
                 .map(EmbeddingsWrap::NonView)
                 .map_err(|err| exceptions::PyIOError::new_err(err.to_string()))?,
         };
@@ -60,7 +63,7 @@ impl PyEmbeddings {
         })
     }
 
-    /// read_fasttext(path,/ lossy)
+    /// read_fasttext(path, /, lossy=False)
     /// --
     ///
     /// Read embeddings in the fasttext format.
@@ -68,7 +71,7 @@ impl PyEmbeddings {
     /// Lossy decoding of the words can be toggled through the lossy param.
     #[staticmethod]
     #[args(lossy = false)]
-    fn read_fasttext(path: &str, lossy: bool) -> PyResult<PyEmbeddings> {
+    fn read_fasttext(path: PathBuf, lossy: bool) -> PyResult<PyEmbeddings> {
         if lossy {
             read_non_fifu_embeddings(path, |r| Embeddings::read_fasttext_lossy(r))
         } else {
@@ -76,16 +79,16 @@ impl PyEmbeddings {
         }
     }
 
-    /// read_floret(path)
+    /// read_floret_text(path)
     /// --
     ///
     /// Read embeddings in the floret text format.
     #[staticmethod]
-    fn read_floret_text(path: &str) -> PyResult<PyEmbeddings> {
+    fn read_floret_text(path: PathBuf) -> PyResult<PyEmbeddings> {
         read_non_fifu_embeddings(path, |r| Embeddings::read_floret_text(r))
     }
 
-    /// read_text(path,/ lossy)
+    /// read_text(path, /, lossy=False)
     /// --
     ///
     /// Read embeddings in text format. This format uses one line per
@@ -96,7 +99,7 @@ impl PyEmbeddings {
     /// Lossy decoding of the words can be toggled through the lossy param.
     #[staticmethod]
     #[args(lossy = false)]
-    fn read_text(path: &str, lossy: bool) -> PyResult<PyEmbeddings> {
+    fn read_text(path: PathBuf, lossy: bool) -> PyResult<PyEmbeddings> {
         if lossy {
             read_non_fifu_embeddings(path, |r| Embeddings::read_text_lossy(r))
         } else {
@@ -104,7 +107,7 @@ impl PyEmbeddings {
         }
     }
 
-    /// read_text_dims(path,/ lossy)
+    /// read_text_dims(path, /, lossy=False)
     /// --
     ///
     /// Read embeddings in text format with dimensions. In this format,
@@ -118,7 +121,7 @@ impl PyEmbeddings {
     /// Lossy decoding of the words can be toggled through the lossy param.
     #[staticmethod]
     #[args(lossy = false)]
-    fn read_text_dims(path: &str, lossy: bool) -> PyResult<PyEmbeddings> {
+    fn read_text_dims(path: PathBuf, lossy: bool) -> PyResult<PyEmbeddings> {
         if lossy {
             read_non_fifu_embeddings(path, |r| Embeddings::read_text_dims_lossy(r))
         } else {
@@ -126,7 +129,7 @@ impl PyEmbeddings {
         }
     }
 
-    /// read_word2vec(path,/ lossy)
+    /// read_word2vec(path, /, lossy=False)
     /// --
     ///
     /// Read embeddings in the word2vec binary format.
@@ -134,7 +137,7 @@ impl PyEmbeddings {
     /// Lossy decoding of the words can be toggled through the lossy param.
     #[staticmethod]
     #[args(lossy = false)]
-    fn read_word2vec(path: &str, lossy: bool) -> PyResult<PyEmbeddings> {
+    fn read_word2vec(path: PathBuf, lossy: bool) -> PyResult<PyEmbeddings> {
         if lossy {
             read_non_fifu_embeddings(path, |r| Embeddings::read_word2vec_binary_lossy(r))
         } else {
@@ -142,6 +145,15 @@ impl PyEmbeddings {
         }
     }
 
+    fn __getitem__(&self, py: Python, word: &str) -> PyResult<Py<PyArray1<f32>>> {
+        let embeddings = self.embeddings.read().unwrap();
+
+        match py.allow_threads(|| embeddings.embedding(word)) {
+            Some(embedding) => Ok(embedding.into_owned().into_pyarray(py).to_owned()),
+            None => Err(exceptions::PyKeyError::new_err("Unknown word and n-grams")),
+        }
+    }
+
     /// Get the model's vocabulary.
     #[getter]
     fn vocab(&self) -> PyResult<PyVocab> {
@@ -154,6 +166,9 @@ impl PyEmbeddings {
         PyStorage::new(self.embeddings.clone())
     }
 
+    /// analogy(self, word1, word2, word3, limit=10, mask=(True, True, True))
+    /// --
+    ///
     /// Perform an anology query.
     ///
     /// This returns words for the analogy query *w1* is to *w2*
@@ -193,7 +208,7 @@ impl PyEmbeddings {
         Self::similarity_results(py, results)
     }
 
-    /// embedding(word,/, default)
+    /// embedding(self, word, /, default=None)
     /// --
     ///
     /// Get the embedding for the given word.
@@ -209,36 +224,34 @@ impl PyEmbeddings {
     #[args(default = "PyEmbeddingDefault::default()")]
     fn embedding(
         &self,
+        py: Python,
         word: &str,
         default: PyEmbeddingDefault,
     ) -> PyResult<Option<Py<PyArray1<f32>>>> {
         let embeddings = self.embeddings.read().unwrap();
-        let gil = pyo3::Python::acquire_gil();
         if let PyEmbeddingDefault::Embedding(array) = &default {
-            if array.as_ref(gil.python()).shape()[0] != embeddings.storage().shape().1 {
+            if array.as_ref(py).shape()[0] != embeddings.storage().shape().1 {
                 return Err(exceptions::PyValueError::new_err(format!(
                     "Invalid shape of default embedding: {}",
-                    array.as_ref(gil.python()).shape()[0]
+                    array.as_ref(py).shape()[0]
                 )));
             }
         }
 
-        if let Some(embedding) = embeddings.embedding(word) {
-            return Ok(Some(
-                embedding.into_owned().into_pyarray(gil.python()).to_owned(),
-            ));
+        if let Some(embedding) = py.allow_threads(|| embeddings.embedding(word)) {
+            return Ok(Some(embedding.into_owned().into_pyarray(py).to_owned()));
         };
         match default {
             PyEmbeddingDefault::Constant(constant) => {
                 let nd_arr = Array1::from_elem([embeddings.storage().shape().1], constant);
-                Ok(Some(nd_arr.into_pyarray(gil.python()).to_owned()))
+                Ok(Some(nd_arr.into_pyarray(py).to_owned()))
             }
             PyEmbeddingDefault::Embedding(array) => Ok(Some(array)),
             PyEmbeddingDefault::None => Ok(None),
         }
     }
 
-    /// embedding(words,/,out)
+    /// embedding_batch(self, words, /, out=None)
     /// --
     ///
     /// Get the embedding for a batch of words. The embeddings are returned
@@ -266,19 +279,35 @@ impl PyEmbeddings {
         (out, present)
     }
 
-    fn embedding_with_norm(&self, word: &str) -> Option<Py<PyTuple>> {
+    /// embedding_with_norm(self, word)
+    /// --
+    ///
+    /// Look up the embedding and norm of a word. The embedding and
+    /// norm are returned as a tuple.
+    fn embedding_with_norm(&self, py: Python, word: &str) -> Option<Py<PyTuple>> {
         let embeddings = self.embeddings.read().unwrap();
 
         use EmbeddingsWrap::*;
-        let embedding_with_norm = match &*embeddings {
+        let embedding_with_norm = py.allow_threads(|| match &*embeddings {
             View(e) => e.embedding_with_norm(word),
             NonView(e) => e.embedding_with_norm(word),
-        };
+        });
 
         embedding_with_norm.map(|e| {
-            let gil = pyo3::Python::acquire_gil();
-            let embedding = e.embedding.into_owned().into_pyarray(gil.python());
-            (embedding, e.norm).into_py(gil.python())
+            let embedding = e.embedding.into_owned().into_pyarray(py);
+            (embedding, e.norm).into_py(py)
+        })
+    }
+
+    /// from_bytes(data)
+    /// --
+    ///
+    /// Deserialize embeddings from `bytes`.
+    #[staticmethod]
+    fn from_bytes(data: &[u8]) -> PyResult<Self> {
+        let embeddings = EmbeddingsWrap::read_embeddings(&mut Cursor::new(data))?;
+        Ok(PyEmbeddings {
+            embeddings: Arc::new(RwLock::new(embeddings)),
         })
     }
 
@@ -303,6 +332,19 @@ impl PyEmbeddings {
         }
     }
 
+    /// quantize(self, n_subquantizers, /, quantizer="pq", n_subquantizer_bits=8, n_iterations=100, n_attempts=1, normalize=True)
+    /// --
+    ///
+    /// Quantize the embeddings with the given hyperparemeters:
+    ///
+    /// * The number of subquantizers
+    /// * The quantizer (``pq``, ``opq``, or ``gausian_opq``).
+    /// * The number of bits per subquantizer.
+    /// * The number of optimization iterations.
+    /// * The number of quantization attempts per iteration.
+    /// * Whether embeddings should be l2-normalized before quantization.
+    ///
+    /// Returns the quantized embeddings.
     #[allow(clippy::too_many_arguments)]
     #[args(
         quantizer = "\"pq\"",
@@ -332,6 +374,9 @@ impl PyEmbeddings {
         )
     }
 
+    /// Set the metadata of the embeddings.
+    ///
+    /// Must be a valid TOML.
     #[setter]
     fn set_metadata(&mut self, metadata: &str) -> PyResult<()> {
         let value = match metadata.parse::<Value>() {
@@ -355,6 +400,25 @@ impl PyEmbeddings {
         Ok(())
     }
 
+    /// to_bytes(self)
+    /// --
+    ///
+    /// Serialize the embeddings to `bytes`.
+    fn to_bytes<'py>(&self, py: Python<'py>) -> PyResult<&'py PyBytes> {
+        let embeddings = self.embeddings.read().unwrap();
+        PyBytes::new_with(
+            py,
+            embeddings.write_embeddings_len(0) as usize,
+            |bytes: &mut [u8]| {
+                let mut cursor = Cursor::new(bytes);
+                embeddings.write_embeddings(&mut cursor)
+            },
+        )
+    }
+
+    /// word_similarity(self, word, /, limit=10)
+    /// --
+    ///
     /// Perform a similarity query.
     #[args(limit = 10)]
     fn word_similarity(&self, py: Python, word: &str, limit: usize) -> PyResult<Vec<PyObject>> {
@@ -373,7 +437,11 @@ impl PyEmbeddings {
         Self::similarity_results(py, results)
     }
 
-    /// Perform a similarity query based on a query embedding.
+    /// embedding_similarity(self, embeddings, /, skip=None, limit=10)
+    /// --
+    ///
+    /// Perform a similarity query based on a query embedding. ``skip``
+    /// specifies the set of words that should never be returned.
     #[args(limit = 10, skip = "Skips(HashSet::new())")]
     fn embedding_similarity(
         &self,
@@ -407,22 +475,15 @@ impl PyEmbeddings {
         Self::similarity_results(py, results)
     }
 
+    /// write(self, path)
+    /// --
+    ///
     /// Write the embeddings to a finalfusion file.
-    fn write(&self, filename: &str) -> PyResult<()> {
-        let f = File::create(filename)?;
-        let mut writer = BufWriter::new(f);
-
+    fn write(&self, path: PathBuf) -> PyResult<()> {
         let embeddings = self.embeddings.read().unwrap();
-
-        use EmbeddingsWrap::*;
-        match &*embeddings {
-            View(e) => e
-                .write_embeddings(&mut writer)
-                .map_err(|err| exceptions::PyIOError::new_err(err.to_string())),
-            NonView(e) => e
-                .write_embeddings(&mut writer)
-                .map_err(|err| exceptions::PyIOError::new_err(err.to_string())),
-        }
+        let f = File::create(path)?;
+        let mut writer = BufWriter::new(f);
+        embeddings.write_embeddings(&mut writer)
     }
 }
 
@@ -536,21 +597,6 @@ impl PyEmbeddings {
     }
 }
 
-#[pyproto]
-impl PyMappingProtocol for PyEmbeddings {
-    fn __getitem__(&self, word: &str) -> PyResult<Py<PyArray1<f32>>> {
-        let embeddings = self.embeddings.read().unwrap();
-
-        match embeddings.embedding(word) {
-            Some(embedding) => {
-                let gil = pyo3::Python::acquire_gil();
-                Ok(embedding.into_owned().into_pyarray(gil.python()).to_owned())
-            }
-            None => Err(exceptions::PyKeyError::new_err("Unknown word and n-grams")),
-        }
-    }
-}
-
 #[pyproto]
 impl PyIterProtocol for PyEmbeddings {
     fn __iter__(slf: PyRefMut<Self>) -> PyResult<PyObject> {
@@ -566,7 +612,7 @@ impl PyIterProtocol for PyEmbeddings {
 }
 
 fn read_embeddings<S>(
-    path: &str,
+    path: impl AsRef<Path>,
     mmap: bool,
 ) -> finalfusion::error::Result<Embeddings<VocabWrap, S>>
 where
@@ -586,16 +632,20 @@ where
     Ok(embeddings)
 }
 
-fn read_non_fifu_embeddings<R, V>(path: &str, read_embeddings: R) -> PyResult<PyEmbeddings>
+fn read_non_fifu_embeddings<R, V>(
+    path: impl AsRef<Path>,
+    read_embeddings: R,
+) -> PyResult<PyEmbeddings>
 where
     R: FnOnce(&mut BufReader<File>) -> finalfusion::error::Result<Embeddings<V, NdArray>>,
     V: Vocab,
     Embeddings<VocabWrap, StorageViewWrap>: From<Embeddings<V, NdArray>>,
 {
-    let f = File::open(path).map_err(|err| {
+    let f = File::open(&path).map_err(|err| {
         exceptions::PyIOError::new_err(format!(
             "Cannot read text embeddings from '{}': {}'",
-            path, err
+            path.as_ref().to_string_lossy(),
+            err
         ))
     })?;
     let mut reader = BufReader::new(f);
@@ -603,7 +653,8 @@ where
     let embeddings = read_embeddings(&mut reader).map_err(|err| {
         exceptions::PyIOError::new_err(format!(
             "Cannot read text embeddings from '{}': {}'",
-            path, err
+            path.as_ref().to_string_lossy(),
+            err
         ))
     })?;
 
